"""Platform components specific to gocept.net."""
from batou.component import Component, platform
from batou.lib.file import File
import batou.lib.service
import os.path


@platform('gocept.net', batou.lib.service.Service)
class UserInit(Component):
    """Register a service with user-init."""

    def configure(self):
        self.executable = self.parent.executable
        self.pidfile = self.parent.pidfile
        if not os.path.isabs(self.pidfile):
            self.pidfile = os.path.join(self.root.workdir, self.pidfile)
        target = '/var/spool/init.d/{0}/{1}'.format(
            self.environment.service_user,
            os.path.basename(self.parent.executable))
        init_source = os.path.join(
            os.path.dirname(__file__), 'resources', 'init.sh')
        self += File(
            target, source=init_source, mode=0o755, leading=True)



from batou.component import Component, HookComponent
from batou.lib.file import File
import os.path


def ServiceCheck(name, **kw):
    if kw.get('nrpe', False):
        return NRPEService(name, **kw)
    return Service(name, **kw)


class Service(HookComponent):

    namevar = 'description'
    key = 'batou.lib.nagios:Service'

    command = None  # path to executable, if relative then to compdir
    args = ''
    notes_url = ''
    servicegroups = 'direct'

    # dependencies are a list of (host_object, check_description)
    depend_on = ()

    @property
    def check_command(self):
        result = self.command
        if self.args:
            result += '!' + self.args
        return result


class NRPEService(Service):

    name = None
    servicegroups = 'nrpe'

    def configure(self):
        super(NRPEService, self).configure()
        if not self.name:
            self.name = self.description.lower().replace(' ', '_')

    @property
    def check_command(self):
        return 'check_nrpe!%s' % self.name

    @property
    def nrpe_command(self):
        return '{} {}'.format(self.command, self.args)


class NagiosServer(Component):
    """Generate a Nagios server config snippet that reflects all active checks.

    The generated config needs to be copied to the Nagios server.

    """

    nagios_cfg = os.path.join(os.path.dirname(__file__),
                              'resources', 'nagios.cfg')

    # Additional static content that will be appended to the generated file.
    static = ''

    def configure(self):
        self.services = list(self.require(Service.key))
        self.services.sort(key=lambda x: (x.host.name, x.description))

        self += File(
            self.expand('nagios-server-{{environment.service_user}}.cfg'),
            source=self.nagios_cfg,
            mode=0o644)


class NRPEHost(Component):
    """Super-component to create an NRPE server config."""

    # XXX gocept-net specific
    nrpe_cfg = os.path.join(os.path.dirname(__file__),
                            'resources', 'nrpe.cfg')

    def configure(self):
        self.services = [
            service for service in self.require(Service.key, host=self.host)
            if isinstance(service, NRPEService)]
        self.services.sort(key=lambda x: x.name)

        self += File(
            self.expand('/etc/nagios/nrpe/local/'
                        '{{environment.service_user}}.cfg'),
            source=self.nrpe_cfg,
            mode=0o644)



import argparse
import batou.init
import batou.local
import batou.remote
import batou.secrets.edit
import batou.update
import logging
import os
import os.path
import pkg_resources
import sys


def main():
    os.chdir(os.path.dirname(sys.argv[0]))
    parser = argparse.ArgumentParser(
        description='batou v%s '
        'multi-(host|component|environment|version|platform) deployment'
        % pkg_resources.resource_string(__name__, 'version.txt'),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='Enable debug mode.')

    subparsers = parser.add_subparsers()

    # INIT
    p = subparsers.add_parser(
        'init',
        help="""\
Initialize batou project in the given directory. If the given directory does
not exist, it will be created.

If no directory is given, the current directory is used.
""")
    p.add_argument('destination')
    p.set_defaults(func=batou.init.main)

    # LOCAL
    p = subparsers.add_parser(
        'local', help=u'Deploy locally.')
    p.add_argument(
        '-p', '--platform', default=None,
        help='Alternative platform to choose. Empty for no platform.')
    p.add_argument(
        '-t', '--timeout', default=None,
        help='Override the environment\'s timeout setting')
    p.add_argument(
        'environment', help='Environment to deploy.',
        nargs='?',
        default='dev',
        type=lambda x: x.replace('.cfg', ''))
    p.add_argument(
        'hostname',
        nargs='?',
        default='localhost',
        help='Host to deploy.')

    p.set_defaults(func=batou.local.main)

    # REMOTE
    p = subparsers.add_parser(
        'remote', help=u'Deploy remotely.')
    p.add_argument(
        'environment', help='Environment to deploy.',
        type=lambda x: x.replace('.cfg', ''))
    p.add_argument(
        '-t', '--timeout', default=None,
        help='Override the environment\'s timeout setting')
    p.add_argument(
        '-D', '--dirty', action='store_true',
        help='Allow deploying with dirty working copy or outgoing changes.')

    p.set_defaults(func=batou.remote.main)

    # SECRETS
    p = subparsers.add_parser(
        'secrets', help="""\
Encrypted secrets file editor utility. Decrypts file,
invokes the editor, and encrypts the file again. If called with a
non-existent file name, a new encrypted file is created.

        Relies on gpg being installed and configured correctly.
""")
    p.add_argument('--editor', '-e', metavar='EDITOR',
                   default=os.environ.get('EDITOR', 'vi'),
                   help='Invoke EDITOR to edit (default: $EDITOR or vi)')
    p.add_argument(
        'environment', help='Environment to edit secrets for.',
        type=lambda x: x.replace('.cfg', ''))
    p.set_defaults(func=batou.secrets.edit.main)

    # UPDATE
    p = subparsers.add_parser(
        'update', help=u'Update the batou version.')
    group = p.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--version', help='Exact version to install.',
        default='')
    group.add_argument(
        '--develop', help='Path to checkout of batou to install in edit mode.',
        default='')
    p.add_argument(
        '--finish', help='(internal)', action='store_true')
    p.set_defaults(func=batou.update.main)

    args = parser.parse_args()

    # Consume global arguments
    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    console = logging.StreamHandler(sys.stdout)
    console.setLevel(level)
    #console.setFormatter(logging.Formatter(fmt='%(message)s'))
    logger = logging.getLogger('batou')
    logger.setLevel(level)
    logger.addHandler(console)

    # Pass over to function
    func_args = dict(args._get_kwargs())
    del func_args['func']
    del func_args['debug']
    args.func(**func_args)



from batou.utils import cmd
import batou.lib.mercurial
import os.path
import pytest


@pytest.fixture(scope='function')
def repos_path(root):
    repos_path = os.path.join(root.environment.workdir_base, 'upstream')
    cmd('mkdir {dir}; cd {dir}; hg init;'
        'touch foo; hg add foo; hg commit -m "foo"'.format(dir=repos_path))
    return repos_path


@pytest.mark.slow
def test_runs_hg_to_clone_repository(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', revision='tip')
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/foo'))
    root.component.deploy()  # trigger verify


@pytest.mark.slow
def test_setting_branch_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component.deploy()
    cmd('cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
        dir=repos_path))
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/bar'))


@pytest.mark.slow
def test_branch_does_switch_branch(root, repos_path):
    cmd('cd {dir}; hg branch bar; hg ci -m "commit branch"'.format(
        dir=repos_path))
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='bar')
    root.component.deploy()
    stdout, stderr = cmd('cd {workdir}/clone; hg branch'.format(
        workdir=root.workdir))
    assert 'bar' == stdout.strip()


@pytest.mark.slow
def test_set_revision_does_not_pull_when_revision_matches(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component += clone
    root.component.deploy()
    revision = clone.current_revision
    clone.revision = revision
    clone.branch = None
    cmd('cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
        dir=repos_path))
    root.component.deploy()
    stdout, stderr = cmd('cd {workdir}/clone; LANG=C hg incoming'.format(
        workdir=root.workdir))
    assert 'changeset:   1' in stdout


@pytest.mark.slow
def test_has_changes_counts_changes_to_tracked_files(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes
    cmd('touch {}/clone/bar'.format(root.workdir))
    cmd('cd {}/clone; hg add bar'.format(root.workdir))
    assert clone.has_changes


@pytest.mark.slow
def test_has_changes_counts_untracked_files_as_changes(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes
    cmd('touch {}/clone/bar'.format(root.workdir))
    assert clone.has_changes


@pytest.mark.slow
def test_clean_clone_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component.deploy()
    cmd('cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
        dir=repos_path))
    root.component.deploy()
    assert os.path.isfile(root.component.map('clone/bar'))


@pytest.mark.slow
def test_changes_lost_on_update_with_incoming(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component.deploy()
    cmd('cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
        dir=repos_path))
    cmd('cd {dir}/clone; echo foobar >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert os.path.exists(root.component.map('clone/bar'))
    assert not open(root.component.map('clone/foo')).read()


@pytest.mark.slow
def test_untracked_files_are_removed_on_update(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component.deploy()
    cmd('cd {dir}/clone; mkdir bar; echo foobar >bar/baz'.format(
        dir=root.workdir))
    root.component.deploy()
    assert not os.path.exists(root.component.map('clone/bar/baz'))


@pytest.mark.slow
def test_changes_lost_on_update_without_incoming(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default')
    root.component.deploy()
    cmd('cd {dir}/clone; echo foobar >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert not open(root.component.map('clone/foo')).read()


@pytest.mark.slow
def test_clean_clone_vcs_update_false_leaves_changes_intact(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target='clone', branch='default', vcs_update=False)
    root.component.deploy()
    cmd('cd {dir}; echo foobar >foo; touch bar; hg addremove; '
        'hg ci -m "commit"'.format(dir=repos_path))
    cmd('cd {dir}/clone; echo asdf >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert 'asdf\n' == open(root.component.map('clone/foo')).read()
    assert not os.path.exists(root.component.map('clone/bar'))



from .component import load_components_from_file
from .host import Host
from .resources import Resources
from .secrets import add_secrets_to_environment_override
from ConfigParser import RawConfigParser
from batou import NonConvergingWorkingSet, UnusedResource
from batou.component import RootComponent
from batou.utils import revert_graph, topological_sort
import batou.c
import batou.vfs
import glob
import logging
import os
import os.path
import pwd
import sys


logger = logging.getLogger(__name__)


class ConfigSection(dict):
    def as_list(self, option):
        result = self[option]
        if ',' in result:
            result = [x.strip() for x in result.split(',')]
        elif '\n' in result:
            result = (x.strip() for x in result.split('\n'))
            result = [x for x in result if x]
        else:
            result = [result]
        return result


class Config(object):
    def __init__(self, path):
        config = RawConfigParser()
        config.optionxform = lambda s: s
        config.read(path)
        self.config = config

    def __contains__(self, section):
        return self.config.has_section(section)

    def __getitem__(self, section):
        if section not in self:
            raise KeyError(section)
        return ConfigSection(
            (x, self.config.get(section, x))
            for x in self.config.options(section))

    def __iter__(self):
        return iter(self.config.sections())

    def get(self, section, default=None):
        try:
            return self[section]
        except KeyError:
            return default


class Environment(object):
    """An environment assigns components to hosts and provides
    environment-specific configuration for components.
    """

    service_user = None
    connect_method = None
    host_domain = None
    branch = None
    update_method = None
    platform = None
    vfs_sandbox = None
    timeout = None

    def __init__(self, name, basedir='.'):
        self.name = name
        self.hosts = {}
        self.resources = Resources()
        self.overrides = {}

        # These are the component classes, decorated with their
        # name.
        self.components = {}
        # These are the components assigned to hosts.
        self.root_components = []

        self.base_dir = os.path.abspath(basedir)
        self.workdir_base = os.path.join(self.base_dir, 'work')

    def load(self):
        # Scan all components
        for filename in sorted(glob.glob(
                os.path.join(self.base_dir, 'components/*/component.py'))):
            self.components.update(load_components_from_file(filename))

        # Load environment configuration
        config_file = os.path.join(
            self.base_dir, 'environments/{}.cfg'.format(self.name))
        if not os.path.isfile(config_file):
            raise ValueError('No such environment "{}"'.format(self.name))

        config = Config(config_file)

        self.load_environment(config)
        for hostname in config.get('hosts', {}):
            components = parse_host_components(
                config['hosts'].as_list(hostname))
            for component, features in components.items():
                self.add_root(component, hostname, features)

        # load overrides
        for section in config:
            if not section.startswith('component:'):
                continue
            root_name = section.replace('component:', '')
            self.overrides.setdefault(root_name, {})
            self.overrides[root_name].update(config[section])

    def load_secrets(self):
        add_secrets_to_environment_override(self)

    def load_environment(self, config):
        environment = config.get('environment', {})
        for key in ['service_user', 'connect_method', 'host_domain',
                    'update_method', 'branch', 'platform', 'timeout']:
            if key not in environment:
                continue
            if getattr(self, key) is not None:
                # Avoid overriding early changes that have already been
                # applied, e.g. by tests.
                continue
            setattr(self, key, environment[key])

        self._set_defaults()

        if 'vfs' in config:
            sandbox = config['vfs']['sandbox']
            sandbox = getattr(batou.vfs, sandbox)(self, config['vfs'])
            self.vfs_sandbox = sandbox

    def _set_defaults(self):
        if self.branch is None:
            self.branch = u'default'
        if self.update_method is None:
            self.update_method = 'pull'

        if self.connect_method is None:
            self.connect_method = 'ssh+sudo'
        if self.service_user is None:
            self.service_user = pwd.getpwuid(os.getuid()).pw_name

        if self.platform is None and self.host_domain:
            self.platform = self.host_domain

        if self.timeout is None:
            self.timeout = 3
        else:
            self.timeout = int(self.timeout)

    # API to instrument environment config loading

    def add_host(self, hostname):
        fqdn = self.normalize_host_name(hostname)
        if fqdn not in self.hosts:
            self.hosts[fqdn] = Host(fqdn)
        return self.hosts[fqdn]

    def add_root(self, component_name, hostname, features=()):
        host = self.add_host(hostname)
        root = RootComponent(component_name, self, host, features)
        self.root_components.append(root)
        return root

    def get_root(self, component_name, hostname):
        host = self.add_host(hostname)
        for root in self.root_components:
            if root.host == host and root.name == component_name:
                return root
        raise KeyError("Component {} not configured for host {}".format(
            component_name, hostname))

    # Deployment API (implements the configure-verify-update cycle)

    def configure(self):
        """Configure all root components.

        Monitor the dependencies between resources and try to reach a stable
        order.

        """
        working_set = set(self.root_components)

        previous_working_sets = []
        while working_set:
            previous_working_sets.append(working_set.copy())
            retry = set()
            exceptions = []
            self.resources.dirty_dependencies.clear()

            for root in working_set:
                try:
                    self.resources.reset_component_resources(root)
                    root.prepare()
                except Exception:
                    exceptions.append((root, sys.exc_info()))
                    retry.add(root)
                    continue

            retry.update(self.resources.dirty_dependencies)
            retry.update(self.resources.unsatisfied_components)

            # Try to find a valid order of the components. If we can't then we
            # have detected a dependency cycle and need to stop.

            self.roots_in_order()

            if (retry in previous_working_sets):
                # We did not manage to improve on our last working set, so we
                # give up.

                # Report all exceptions we got in the last run.
                for root, e in exceptions:
                    logger.error('', exc_info=e)

                # If any resources were required but not provided at least
                # once we report this as well.
                if self.resources.unsatisfied:
                    logger.error('Unsatisfied resources: %s' %
                                 ', '.join(self.resources.unsatisfied))

                if exceptions:
                    logger.error('\nThe following exceptions occurred '
                                 '(see tracebacks above):')
                for root, e in exceptions:
                    logger.error('%s: %r', root.name, e[1])
                raise NonConvergingWorkingSet(retry)

            working_set = retry

        # We managed to converge on a working set. However, some resource were
        # provided but never used. We're rather picky here and report this as
        # an error.
        if self.resources.unused:
            raise UnusedResource(self.resources.unused)

    def roots_in_order(self, host=None):
        """Return a list of root components sorted by their resource
        dependencies, filtered by host if specified.

        Raises ValueError if component dependencies have cycles.

        """
        dependencies = self.resources.get_dependency_graph()
        # Complete the graph with components that do not have any dependencies
        # (yet)
        for root in self.root_components:
            if root not in dependencies:
                dependencies[root] = set()
        roots = topological_sort(revert_graph(dependencies))
        if host is not None:
            host = self.get_host(host)
            roots = filter(lambda x: x.host is host, roots)
        return roots

    def normalize_host_name(self, hostname):
        """Ensure the given host name is an FQDN for this environment."""
        if not self.host_domain or hostname.endswith(self.host_domain):
            return hostname
        else:
            return '%s.%s' % (hostname, self.host_domain)

    def get_host(self, hostname):
        return self.hosts[self.normalize_host_name(hostname)]

    def map(self, path):
        if self.vfs_sandbox:
            return self.vfs_sandbox.map(path)
        return path


def parse_host_components(components):
    """Parse a component list as given in an environment config for a host
    into a mapping of compoment -> features.

    Expected syntax:

    component[:feature], component[:feature]

    """
    result = {}
    for name in components:
        name = name.strip()
        if ':' in name:
            name, feature = name.split(':', 1)
        else:
            feature = None
        result.setdefault(name, [])
        if feature:
            result[name].append(feature)
    return result



from . import remote_core
from .environment import Environment
from .utils import notify, cmd
import execnet
import logging
import os
import os.path
import subprocess
import sys
import tempfile

# XXX reset

logger = logging.getLogger('batou.remote')


def main(environment, timeout, dirty):
    environment = Environment(environment)
    environment.load()
    if timeout is not None:
        environment.timeout = timeout
    if not dirty:
        check_clean_hg_repository()
    environment.load_secrets()
    environment.configure()

    deployment = RemoteDeployment(environment, dirty)
    try:
        deployment()
    except Exception:
        logger.error('', exc_info=True)
    else:
        notify('Deployment finished',
               '{} was deployed successfully.'.format(environment.name))


def check_clean_hg_repository():
    # Safety belt that we're acting on a clean repository.
    try:
        status, _ = cmd('hg -q stat', silent=True)
    except RuntimeError:
        logger.error('Unable to check repository status. '
                     'Is there an HG repository here?')
        sys.exit(1)
    else:
        status = status.strip()
        if status.strip():
            logger.error("""\
Your repository has uncommitted changes.

I am refusing to deploy in this situation as the results will be unpredictable.
Please commit and push first.
""")
            logger.error(status)
            sys.exit(1)

    try:
        cmd('hg -q outgoing -l 1', acceptable_returncodes=[1])
    except RuntimeError:
        logger.error("""\
Your repository has outgoing changes.

I am refusing to deploy in this situation as the results will be unpredictable.
Please push first.
""")
        sys.exit(1)


class RemoteDeployment(object):

    def __init__(self, environment, dirty):
        self.environment = environment
        self.dirty = dirty

        self.upstream = cmd('hg showconfig paths')[0].split('\n')[0].strip()
        assert self.upstream.startswith('paths.default')
        self.upstream = self.upstream.split('=')[1]

        self.repository_root = subprocess.check_output(['hg', 'root']).strip()
        self.deployment_base = os.path.relpath(
            self.environment.base_dir, self.repository_root)
        assert (self.deployment_base == '.' or
                self.deployment_base[0] not in ['.', '/'])

    def __call__(self):
        remotes = {}
        for host in self.environment.hosts.values():
            remote = RemoteHost(host, self)
            remotes[host] = remote
            remote.connect()

        # Bootstrap and get channel to batou environment
        for remote in remotes.values():
            remote.start()

        for root in self.environment.roots_in_order():
            remote = remotes[root.host]
            remote.deploy_component(root)

        for remote in remotes.values():
            remote.gateway.exit()


class RPCWrapper(object):

    def __init__(self, host):
        self.host = host

    def __getattr__(self, name):
        def call(*args, **kw):
            logger.debug('rpc {}: {}(*{}, **{})'.format
                         (self.host.host.fqdn, name, args, kw))
            self.host.channel.send((name, args, kw))
            result = self.host.channel.receive()
            logger.debug('result: {}'.format(result))
            try:
                result[0]
            except (TypeError, IndexError):
                pass
            else:
                if result[0] == 'batou-remote-core-error':
                    logger.error(result[1])
                    raise RuntimeError('Remote exception encountered.')
            return result
        return call

    def send_file(self, local, remote):
        self.host.channel.send(('send_file', [remote], {}))
        with open(local, 'r') as f:
            while True:
                data = f.read(64*1024)
                if not data:
                    self.host.channel.send(('finish', ''))
                    break
                else:
                    self.host.channel.send((None, data))
        result = self.host.channel.receive()
        logger.debug('result: {}'.format(result))
        try:
            result[0]
        except TypeError:
            pass
        else:
            if result[0] == 'batou-remote-core-error':
                logger.error(result[1])
                raise RuntimeError('Remote exception encountered.')
        assert result == 'OK'


class RemoteHost(object):

    gateway = None

    def __init__(self, host, deployment):
        self.host = host
        self.deployment = deployment
        self.rpc = RPCWrapper(self)

    def connect(self, interpreter='python2.7'):
        if not self.gateway:
            logger.info('{}: connecting'.format(self.host.fqdn))
        else:
            logger.info('{}: reconnecting'.format(self.host.fqdn))
            self.gateway.exit()

        if self.deployment.environment.connect_method == 'ssh+sudo':
            sudo = 'sudo -u {} '.format(
                self.deployment.environment.service_user)
        else:
            sudo = ''

        self.gateway = execnet.makegateway(
            "ssh={}//python={} {}".format(
                self.host.fqdn,
                sudo,
                interpreter))
        self.channel = self.gateway.remote_exec(remote_core)

    def start(self):
        logger.info('{}: bootstrapping'.format(self.host.fqdn))
        self.rpc.lock()
        env = self.deployment.environment

        remote_repository = self.rpc.ensure_repository()
        self.remote_base = os.path.join(
            remote_repository, self.deployment.deployment_base)

        if env.update_method == 'pull':
            self.rpc.pull_code(
                upstream=self.deployment.upstream)
        elif env.update_method == 'bundle':
            heads = self.rpc.current_heads()
            fd, bundle_file = tempfile.mkstemp()
            os.close(fd)
            bases = ' '.join('--base {}'.format(x) for x in heads)
            cmd('hg -qy bundle {} {}'.format(bases, bundle_file),
                acceptable_returncodes=[0, 1])
            have_changes = os.stat(bundle_file).st_size > 0
            self.rpc.send_file(
                bundle_file, remote_repository + '/batou-bundle.hg')
            os.unlink(bundle_file)
            if have_changes:
                self.rpc.unbundle_code()
        else:
            raise ValueError(
                'unsupported update method: {}'.format(env.update_method))

        remote_id = self.rpc.update_working_copy(env.branch)
        local_id, _ = cmd('hg id -i')
        if self.deployment.dirty:
            local_id = local_id.replace('+', '')
        local_id = local_id.strip()
        if remote_id != local_id:
            raise RuntimeError(
                'Working copy parents differ. Local: {} Remote: {}'.format(
                    local_id, remote_id))

        self.rpc.build_batou(self.deployment.deployment_base)

        # Now, replace the basic interpreter connection, with a "real" one that
        # has all our dependencies installed.
        self.connect(self.remote_base + '/.batou/bin/python')

        self.rpc.setup_deployment(
            self.deployment.deployment_base,
            env.name,
            self.host.fqdn,
            env.overrides)

    def deploy_component(self, component):
        logger.info('Deploying {}/{}'.format(self.host.fqdn, component.name))
        self.rpc.deploy(component.name)



from batou.environment import Environment
import os.path


def test_service_early_resource():
    env = Environment(
        'dev',
        os.path.dirname(__file__)+'/fixture/service_early_resource')
    env.load()
    env.configure()
    assert env.resources.get('zeo') == ['127.0.0.1:9000']



from batou.component import Component
import batou
import batou.utils
import os.path
import urlparse
import requests
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve


class Download(Component):

    namevar = 'uri'

    target = None  # Filename where the download will be stored.
    checksum = None
    requests_kwargs = None

    def configure(self):
        if not self.target:
            self.target = urlparse.urlsplit(self.uri).path.split('/')[-1]
        if not self.target:
            raise KeyError('No target is given and the URI does not allow '
                           'deriving a filename.')
        if not self.checksum:
            raise ValueError('No checksum given.')
        self.checksum_function, self.checksum = self.checksum.split(':')

    def verify(self):
        if not os.path.exists(self.target):
            raise batou.UpdateNeeded()
        if self.checksum != batou.utils.hash(self.target,
                                             self.checksum_function):
            raise batou.UpdateNeeded()

    def update(self):
        scheme = urlparse.urlsplit(self.uri)[0]
        if scheme in ['http', 'https']:
            self._update_requests()
        else:
            self._update_urllib()

        target_checksum = batou.utils.hash(self.target, self.checksum_function)
        assert self.checksum == target_checksum, '''\
Checksum mismatch!
expected: %s
got: %s''' % (self.checksum, target_checksum)

    def _update_requests(self):
        r = requests.get(
            self.uri,
            **(self.requests_kwargs if self.requests_kwargs else {}))
        r.raise_for_status()

        with open(self.target, 'wb') as fd:
            for chunk in r.iter_content(4*1024**2):
                fd.write(chunk)

    def _update_urllib(self):
        path, headers = urlretrieve(self.uri, self.target)
        assert path == self.target



from batou.utils import cmd
import batou.lib.git
import os.path
import pytest


@pytest.mark.slow
def test_runs_git_to_clone_repository(root):
    repos_path = os.path.join(root.environment.workdir_base, 'upstream')
    cmd('mkdir {dir}; cd {dir}; git init;'
        'touch foo; git add foo;'
        'git commit --author="Jenkins <jenkins@example.com>" -am "foo"'.format(
            dir=repos_path))
    root.component += batou.lib.git.Clone(repos_path, target='clone')
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/foo'))
    root.component.deploy()  # trigger verify


@pytest.mark.slow
def test_git_offers_bbb_api(root):
    repos_path = os.path.join(root.environment.workdir_base, 'upstream')
    cmd('mkdir {dir}; cd {dir}; git init;'
        'touch foo; git add foo;'
        'git commit --author="Jenkins <jenkins@example.com>" -am "foo"'.format(
            dir=repos_path))
    root.component += batou.lib.git.Clone('clone', source=repos_path)
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/foo'))



from zest.releaser.utils import system
import logging
import os

logger = logging.getLogger(__name__)


def update_requirements(data):
    os.chdir(data['workingdir'])
    logger.info('Running buildout to update requirements.txt.')
    system('bin/buildout')
    logger.info('Committing requirements.txt.')
    system('hg commit -v -m "Update requirements.txt"')



from batou import remote_core
import inspect
import mock
import os.path
import pytest


def test_deployment_and_channel_exist_as_names():
    assert remote_core.channel is None
    assert remote_core.deployment is None


def test_lock():
    remote_core.lock()


def test_cmd():
    result = remote_core.cmd('echo "asdf"')
    assert result == "asdf\n"


@pytest.fixture
def mock_remote_core(monkeypatch, tmpdir):
    # Suppress active actions in the remote_core module
    monkeypatch.setattr(remote_core, 'cmd', mock.Mock())
    monkeypatch.setattr(remote_core, 'target_directory', mock.Mock())
    remote_core.target_directory.return_value = str(tmpdir)


def test_update_code_existing_target(mock_remote_core):
    remote_core.ensure_repository()
    remote_core.pull_code('http://bitbucket.org/gocept/batou')
    remote_core.update_working_copy('default')

    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next().startswith('hg init /')
    assert calls.next() == 'hg pull http://bitbucket.org/gocept/batou'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'
    assert remote_core.cmd.call_count == 4


def test_update_code_new_target(mock_remote_core):
    remote_core.target_directory.return_value += '/foo'

    remote_core.ensure_repository()
    remote_core.pull_code('http://bitbucket.org/gocept/batou')
    remote_core.update_working_copy('default')

    assert remote_core.cmd.call_count == 4
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next() == 'hg init {}'.format(
        remote_core.target_directory())
    assert calls.next() == 'hg pull http://bitbucket.org/gocept/batou'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'


def test_bundle_shipping(mock_remote_core):
    remote_core.target_directory.return_value += '/foo'

    remote_core.ensure_repository()
    remote_core.cmd.return_value = """\
changeset: 371:revision-a
nsummary:fdsa

changeset: 372:revision-b
"""
    heads = remote_core.current_heads()
    assert heads == ['revision-a', 'revision-b']
    remote_core.unbundle_code()
    remote_core.update_working_copy('default')

    assert remote_core.cmd.call_count == 6
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next() == 'hg init {}'.format(
        remote_core.target_directory())
    assert calls.next() == 'hg id -i'
    assert calls.next() == 'hg heads'
    assert calls.next() == 'hg -y unbundle batou-bundle.hg'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'


def test_build_batou_fresh_install(mock_remote_core):
    remote_core.build_batou('.')
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 1
    assert calls.next() == './batou --help'


def test_build_batou_virtualenv_exists(mock_remote_core):
    os.mkdir(remote_core.target_directory() + '/bin')
    open(remote_core.target_directory() + '/bin/python2.7', 'w')
    remote_core.build_batou('.')
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 1
    assert calls.next() == './batou --help'


def test_expand_deployment_base():
    assert (remote_core.target_directory() ==
            os.path.expanduser('~/deployment'))


def test_deploy_component(monkeypatch):
    monkeypatch.setattr(remote_core, 'deployment', mock.Mock())
    remote_core.deploy('foo')
    assert remote_core.deployment.deploy.call_count == 1


class DummyChannel(object):

    _closed = False

    def __init__(self):
        self.receivequeue = []
        self.sendqueue = []

    def isclosed(self):
        return self._closed

    def receive(self, timeout=None):
        result = self.receivequeue.pop(0)
        if not self.receivequeue:
            self._closed = True
        return result

    def send(self, item):
        self.sendqueue.append(item)


@pytest.fixture
def remote_core_mod():
    channel = DummyChannel()

    local_namespace = dict(channel=channel, __name__='__channelexec__')

    remote_core_mod = compile(
        inspect.getsource(remote_core),
        inspect.getsourcefile(remote_core),
        'exec')

    def run():
        exec remote_core_mod in local_namespace

    return (channel, run)


def test_channelexec_already_closed(remote_core_mod):
    channel, run = remote_core_mod
    channel._closed = True
    run()
    assert channel.receivequeue == []
    assert channel.sendqueue == []


def test_channelexec_echo_cmd(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('echo "asdf"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == ['asdf\n']


def test_channelexec_multiple_echo_cmds(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('echo "asdf1"',), {}))
    channel.receivequeue.append(('cmd', ('echo "asdf2"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == ['asdf1\n', 'asdf2\n']


def test_channelexec_handle_exception(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('fdjkahfkjdasbfda',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    response = channel.sendqueue[0]
    assert response[0] == 'batou-remote-core-error'
    assert ('CalledProcessError: Command \'[\'fdjkahfkjdasbfda\']\' returned '
            'non-zero exit status 127' in response[1])



# -*- coding: utf-8 -*-
from batou.template import TemplateEngine
import collections
import jinja2
import mock
import os.path
import pytest


Server = collections.namedtuple('Server', ['name', 'address'])


sample_dict = dict(
    host=mock.Mock(),
    config=dict(),
    servers=[Server('s1', '1.2.3.4'), Server('s2', '2.3.4.5')],
    hello='world',
    hello2=u'wörld')

fixture = '%s/fixture/template' % os.path.dirname(__file__)


def test_unknown_template_engine():
    with pytest.raises(NotImplementedError):
        TemplateEngine.get('foo')


def test_abstract_engine():
    engine = TemplateEngine()
    with pytest.raises(NotImplementedError):
        engine._render_template_file("asdf", {})
    with pytest.raises(NotImplementedError):
        engine.expand("asdf", {})


def test_jinja2_template_str():
    tmpl = TemplateEngine.get('jinja2')
    assert 'hello world' == tmpl.expand('hello {{hello}}', sample_dict)


def test_jinja2_template_file():
    tmpl = TemplateEngine.get('jinja2')
    filename = '{}/haproxy.cfg'.format(fixture)
    result = tmpl.template(filename + '.jinja2', sample_dict)
    with open(filename) as ref:
        assert ref.read() == result


def test_jinja2_unknown_variable_should_fail():
    tmpl = TemplateEngine.get('jinja2')
    with pytest.raises(jinja2.UndefinedError):
        tmpl.expand('unknown variable {{foo}}', {})


def test_jinja2_umlaut_variables():
    tmpl = TemplateEngine.get('jinja2')
    assert u'hello wörld' == tmpl.expand('hello {{hello2}}', sample_dict)



from batou.resources import Resources
import mock


def test_regression_reset_works_if_provided_but_not_required():
    resources = Resources()
    root = mock.Mock()
    resources.provide(root, mock.sentinel.key, 'asdf')
    resources.reset_component_resources(root)


def test_reset_marks_depending_components_as_dirty():
    resources = Resources()
    root = mock.Mock()
    resources.provide(
        root, mock.sentinel.key, 'asdf')
    resources.require(
        root, mock.sentinel.key)
    assert resources.dirty_dependencies == set()
    resources.reset_component_resources(root)
    assert resources.dirty_dependencies == set([root])



# This code must not cause non-stdlib imports to support self-bootstrapping.


class UpdateNeeded(Exception):
    """A component requires an update."""
    pass


class UnusedResource(Exception):
    """A provided resource was never used."""


class NonConvergingWorkingSet(Exception):
    """A working set did not converge."""



from .environment import Environment
from .utils import notify, locked, CmdExecutionError
from batou import NonConvergingWorkingSet
import logging
import sys


logger = logging.getLogger(__name__)


def main(environment, hostname, platform, timeout):
    with locked('.batou-lock'):
        try:
            environment = Environment(environment)
            environment.load()
            if timeout is not None:
                environment.timeout = timeout
            if platform is not None:
                environment.platform = platform
            environment.load_secrets()
            environment.configure()
            for root in environment.roots_in_order(host=hostname):
                root.component.deploy()
        except (CmdExecutionError, NonConvergingWorkingSet):
            # this has already been reported by the component itself.
            notify('Deployment failed',
                   '{}:{} encountered an error.'.format(
                       environment.name, hostname))
            sys.exit(1)
        except Exception:
            logger.error('', exc_info=True)
            notify('Deployment failed',
                   '{}:{} encountered an error.'.format(
                       environment.name, hostname))
            sys.exit(1)
        else:
            notify('Deployment finished',
                   '{}:{} was deployed successfully.'.format(
                       environment.name, hostname))



from batou import UpdateNeeded
from batou.component import Component, HookComponent, platform
from batou.lib.file import File
import os


class CronJob(HookComponent):

    namevar = 'command'
    key = 'batou.lib.cron:CronJob'

    args = ''
    timing = None
    logger = None

    def format(self):
        if self.timing is None:
            raise ValueError('Required timing value missing from cron job %r.'
                             % self.command)
        line = self.expand(
            '{{component.timing}} {{component.command}} {{component.args}}')
        if self.logger:
            line += self.expand(' 2>&1 | logger -t {{component.logger}}')
        return line


def ignore_comments(data):
    lines = data.splitlines()
    lines = filter(lambda x: not x.startswith('#'), lines)
    return '\n'.join(lines)


class CronTab(Component):

    crontab_template = os.path.join(
        os.path.dirname(__file__), 'resources', 'crontab')
    mailto = None

    def configure(self):
        self.jobs = self.require(CronJob.key, host=self.host)
        self.jobs.sort(key=lambda job: job.command + ' ' + job.args)
        self.crontab = File(
            'crontab', source=self.crontab_template)
        self += self.crontab


class InstallCrontab(Component):

    def configure(self):
        self.crontab = self.parent.crontab

    def verify(self):
        try:
            current, _ = self.cmd('crontab -l', silent=True)
        except:
            current = ''
        current = ignore_comments(current)
        new = ignore_comments(self.crontab.content)
        if new != current:
            raise UpdateNeeded()

    def update(self):
        self.cmd(self.expand('crontab {{component.crontab.path}}'))


@platform('gocept.net', CronTab)
class FCInstallCrontab(InstallCrontab):
    pass


@platform('debian', CronTab)
class DebianInstallCrontab(InstallCrontab):
    pass
